@@ -1540,22 +1540,16 @@ FUN(PDFDocument_addEmbeddedFile)(JNIEnv *env, jobject self, jstring jfilename, j
 }
 
 JNIEXPORT jstring JNICALL
-FUN(PDFDocument_getEmbeddedFileParams)(JNIEnv *env, jobject self, jobject jfs)
-{
-	return FUN(PDFDocument_getFileSpecParams)(env, self, jfs);
-}
-
-JNIEXPORT jstring JNICALL
-FUN(PDFDocument_getFileSpecParams)(JNIEnv *env, jobject self, jobject jfs)
+FUN(PDFDocument_getFilespecParams)(JNIEnv *env, jobject self, jobject jfs)
 {
 	fz_context *ctx = get_context(env);
 	pdf_obj *fs = from_PDFObject_safe(env, jfs);
-	pdf_embedded_file_params params;
+	pdf_filespec_params params;
 	jstring jfilename = NULL;
 	jstring jmimetype = NULL;
 
 	fz_try(ctx)
-		pdf_get_embedded_file_params(ctx, fs, &params);
+		pdf_get_filespec_params(ctx, fs, &params);
 	fz_catch(ctx)
 		jni_rethrow(env, ctx);
 
