@@ -471,11 +471,12 @@
 
 
 static fz_stext_block *
-add_grid_block(fz_context *ctx, fz_stext_page *page, fz_stext_block **first_block, fz_stext_block **last_block)
+add_grid_block(fz_context *ctx, fz_stext_page *page, fz_stext_block **first_block, fz_stext_block **last_block, int id)
 {
 	fz_stext_block *block = fz_pool_alloc(ctx, page->pool, sizeof(**first_block));
 	memset(block, 0, sizeof(*block));
 	block->type = FZ_STEXT_BLOCK_GRID;
+	block->id = id;
 	block->bbox = fz_empty_rect; /* Fixes bug 703267. */
 	block->next = *first_block;
 	if (*first_block)
@@ -589,6 +590,7 @@ add_struct_block_before(fz_context *ctx, fz_stext_block *before, fz_stext_page *
 	/* Now make our new struct block and insert it. */
 	block = fz_pool_alloc(ctx, page->pool, sizeof(*block));
 	block->type = FZ_STEXT_BLOCK_STRUCT;
+	block->id = 0; /* Safe because we only work on single pages */
 	block->bbox = fz_empty_rect; /* Fixes bug 703267. */
 	insert_block_before(block, before, page, parent);
 
@@ -625,11 +627,16 @@ div_list_push(fz_context *ctx, div_list *div, int left, int weak, float pos)
 	{
 		if (div->list[i].pos > pos)
 			break;
-		else if (div->list[i].pos == pos && div->list[i].left == left)
+		else if (div->list[i].pos == pos)
 		{
-			div->list[i].freq++;
-			div->list[i].weak &= weak;
-			return;
+			if (div->list[i].left == left)
+			{
+				div->list[i].freq++;
+				div->list[i].weak &= weak;
+				return;
+			}
+			if (div->list[i].left == 0)
+				break;
 		}
 	}
 
@@ -1156,6 +1163,7 @@ typedef struct
 	fz_stext_grid_positions *xpos;
 	fz_stext_grid_positions *ypos;
 	fz_rect bounds;
+	int has_background;
 } grid_walker_data;
 
 static cell_t *
@@ -1341,6 +1349,12 @@ find_grid_pos(fz_context *ctx, grid_walker_data *gd, int row, float x, int inacc
 		else if (x <= pos->list[i].max)
 		{
 			/* We are in the range for the ith divider. */
+			if (i == 0 || i == pos->len-1)
+			{
+				/* Never move the outermost pos in, because they have been
+				 * calculated to be just big enough already. */
+				return i;
+			}
 			if (pos->list[i].reinforcement == 0)
 			{
 				/* If we've not been reinforced before, reinforce now. */
@@ -1424,8 +1438,6 @@ add_h_line(fz_context *ctx, grid_walker_data *gd, float x0, float x1, float y0,
 	int yidx = find_grid_pos(ctx, gd, 1, y, 0);
 	int i;
 
-	assert(start > 0 && end > 0);
-
 	for (i = start; i < end; i++)
 		get_cell(gd->cells, i, yidx)->h_line++;
 }
@@ -1469,6 +1481,160 @@ add_hv_line(fz_context *ctx, grid_walker_data *gd, float x0, float x1, float y0,
 	}
 }
 
+/* Many PDFs with tables in them start by clearing the area of the table to white
+ * by one big rectangle. Let's try and spot that. */
+enum
+{
+	KEEP_SEARCHING = 0,
+	BACKGROUND_FOUND = 1,
+	NON_BACKGROUND_FOUND = 2
+};
+
+static fz_rect
+whittle_background(fz_context *ctx, fz_stext_block *block0, fz_rect obounds, fz_rect bounds)
+{
+	fz_stext_block *block;
+
+	for (block = block0; block != NULL; block = block->next)
+	{
+		if (block->type == FZ_STEXT_BLOCK_STRUCT)
+		{
+			if (!block->u.s.down)
+				continue;
+			obounds = whittle_background(ctx, block->u.s.down->first_block, obounds, bounds);
+			continue;
+		}
+		else if (block->type == FZ_STEXT_BLOCK_VECTOR ||
+			block->type == FZ_STEXT_BLOCK_TEXT)
+		{
+			fz_rect r = block->bbox;
+
+			if (r.y1 >= bounds.y0 && r.y0 <= bounds.y1)
+			{
+				if (r.x1 <= bounds.x0)
+					obounds.x0 = fz_max(obounds.x0, r.x1);
+				if (r.x0 >= bounds.x1)
+					obounds.x1 = fz_min(obounds.x1, r.x0);
+			}
+			if (r.x1 >= bounds.x0 && r.x0 <= bounds.x1)
+			{
+				if (r.y1 <= bounds.y0)
+					obounds.y0 = fz_max(obounds.y0, r.y1);
+				if (r.y0 >= bounds.y1)
+					obounds.y1 = fz_min(obounds.y1, r.y0);
+			}
+		}
+	}
+
+	return obounds;
+}
+
+/* Can we find a fill that covers bounds, but not obounds, before any content impinges on bounds? */
+static int
+background_hunt(fz_context *ctx, grid_walker_data *gd, fz_stext_block *block0, fz_rect bounds, fz_rect obounds)
+{
+	int found;
+	fz_stext_block *block;
+
+	for (block = block0; block != NULL; block = block->next)
+	{
+		if (block->type == FZ_STEXT_BLOCK_STRUCT)
+		{
+			if (!block->u.s.down)
+				continue;
+			found = background_hunt(ctx, gd, block->u.s.down->first_block, bounds, obounds);
+			if (found)
+				return found;
+			continue;
+		}
+		else if (block->type == FZ_STEXT_BLOCK_VECTOR)
+		{
+			fz_rect r = block->bbox;
+			fz_rect s;
+
+			r = fz_intersect_rect(r, gd->bounds);
+			s = fz_intersect_rect(r, bounds);
+
+			if (!fz_is_valid_rect(s))
+				continue;
+
+			/* Only rectangular && non-stroked blocks can possibly be background
+			 * fills. */
+			if ((block->u.v.flags & FZ_STEXT_VECTOR_IS_RECTANGLE) == 0 ||
+				(block->u.v.flags & FZ_STEXT_VECTOR_IS_STROKED) != 0)
+			{
+				if (r.x0 < obounds.x0 || r.x1 > obounds.x1 || r.y0 < obounds.y0 || r.y1 > obounds.y1)
+				{
+					/* If this extends outside the outer bounds of the table, then maybe it's
+					 * a background fill? Just continue. */
+				}
+				/* Otherwise, this might be content. Stop searching for the background fill. */
+				return NON_BACKGROUND_FOUND;
+			}
+
+			/* Rectangle */
+			if (r.x0 <= bounds.x0 && r.x1 >= bounds.x1 && r.y0 <= bounds.y0 && r.y1 >= bounds.y1)
+			{
+				/* This rectangle covers the entire table. */
+				if (r.x0 < obounds.x0 || r.x1 > obounds.x1 || r.y0 < obounds.y0 || r.y1 > obounds.y1)
+				{
+					/* But it's too big for the table! Maybe it's a page fill? Just ignore it. */
+					continue;
+				}
+				/* So let's extend the table. */
+				gd->xpos->list[0].min = gd->xpos->list[0].pos = gd->xpos->list[0].max = r.x0;
+				gd->xpos->list[gd->xpos->len-1].min = gd->xpos->list[gd->xpos->len-1].pos = gd->xpos->list[gd->xpos->len-1].max = r.x1;
+				gd->ypos->list[0].min = gd->ypos->list[0].pos = gd->ypos->list[0].max = r.y0;
+				gd->ypos->list[gd->ypos->len-1].min = gd->ypos->list[gd->ypos->len-1].pos = gd->ypos->list[gd->ypos->len-1].max = r.y1;
+
+				/* And stop searching */
+				return BACKGROUND_FOUND;
+			}
+			else
+			{
+				/* No background for the table found. */
+				return NON_BACKGROUND_FOUND;
+			}
+		}
+		else if (block->type == FZ_STEXT_BLOCK_TEXT)
+		{
+			fz_rect r = block->bbox;
+			fz_rect s;
+
+			s = fz_intersect_rect(r, gd->bounds);
+			s = fz_intersect_rect(s, bounds);
+
+			if (!fz_is_valid_rect(s))
+				continue;
+
+			return NON_BACKGROUND_FOUND;
+		}
+	}
+
+	return KEEP_SEARCHING;
+}
+
+static int
+walk_for_background(fz_context *ctx, grid_walker_data *gd, fz_stext_block *block0)
+{
+	fz_rect bounds, obounds;
+
+	/* So the smallest covering for our table needs to be this big. */
+	bounds.x0 = gd->xpos->list[0].min;
+	bounds.x1 = gd->xpos->list[gd->xpos->len-1].max;
+	bounds.y0 = gd->ypos->list[0].min;
+	bounds.y1 = gd->ypos->list[gd->ypos->len-1].max;
+
+	obounds = fz_infinite_rect;
+
+	/* What is the biggest margin around this table?
+	 * Start with an infinite box, and whittle it down by any content that
+	 * isn't covered by the bounds. */
+	obounds = whittle_background(ctx, block0, obounds, bounds);
+
+	return background_hunt(ctx, gd, block0, bounds, obounds);
+}
+
 /* Shared internal routine with stext-boxer.c  */
 fz_rect fz_collate_small_vector_run(fz_stext_block **blockp);
 
@@ -2060,6 +2226,8 @@ move_contained_content(fz_context *ctx, fz_stext_page *page, fz_stext_struct *de
 					{
 						newblock = fz_pool_alloc(ctx, page->pool, sizeof(fz_stext_block));
 						insert_block_before(newblock, before, page, dest);
+						newblock->type = FZ_STEXT_BLOCK_TEXT;
+						newblock->id = 0;
 						newblock->u.t.flags = block->u.t.flags;
 						assert(before == newblock->next);
 					}
@@ -2129,6 +2297,8 @@ move_contained_content(fz_context *ctx, fz_stext_page *page, fz_stext_struct *de
 
 							/* Add the block onto our target list */
 							insert_block_before(newblock, before, page, dest);
+							newblock->type = FZ_STEXT_BLOCK_TEXT;
+							newblock->id = 0;
 							newblock->u.t.flags = block->u.t.flags;
 							before = newblock->next;
 						}
@@ -2436,6 +2606,11 @@ transcribe_table(fz_context *ctx, grid_walker_data *gd, fz_stext_page *page, fz_
 	/* Convert to 'before' */
 	if (insert.block)
 		insert.block = insert.block->next;
+	else
+	{
+		insert.block = parent ? parent->first_block : page->first_block;
+		insert.parent = parent;
+	}
 
 	/* Make table */
 	table = add_struct_block_before(ctx, insert.block, page, insert.parent, FZ_STRUCTURE_TABLE, "Table");
@@ -2539,7 +2714,7 @@ transcribe_table(fz_context *ctx, grid_walker_data *gd, fz_stext_page *page, fz_
 		fz_stext_block *block;
 		fz_stext_grid_positions *xps2 = copy_grid_positions_to_pool(ctx, page, gd->xpos);
 		fz_stext_grid_positions *yps2 = copy_grid_positions_to_pool(ctx, page, gd->ypos);
-		block = add_grid_block(ctx, page, &table->first_block, &table->last_block);
+		block = add_grid_block(ctx, page, &table->first_block, &table->last_block, table->up->id);
 		block->u.b.xs = xps2;
 		block->u.b.ys = yps2;
 		block->bbox.x0 = block->u.b.xs->list[0].pos;
@@ -2912,6 +3087,7 @@ find_table_within_bounds(fz_context *ctx, grid_walker_data *gd, fz_stext_block *
 	div_list xs = { 0 };
 	div_list ys = { 0 };
 	int failed = 1;
+	int bg;
 
 	fz_try(ctx)
 	{
@@ -2929,6 +3105,10 @@ find_table_within_bounds(fz_context *ctx, grid_walker_data *gd, fz_stext_block *
 		gd->ypos = make_table_positions(ctx, &ys, bounds.y0, bounds.y1);
 		gd->cells = new_cells(ctx, gd->xpos->len, gd->ypos->len);
 
+		bg = walk_for_background(ctx, gd, content);
+		if (bg == BACKGROUND_FOUND)
+			gd->has_background = 1;
+
 		/* Walk the content looking for grid lines. These
 		 * lines refine our positions. */
 		walk_grid_lines(ctx, gd, content);
@@ -2968,7 +3148,7 @@ find_table_within_bounds(fz_context *ctx, grid_walker_data *gd, fz_stext_block *
 }
 
 static int
-do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent)
+do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent, int *has_background, fz_rect top_bounds)
 {
 	fz_stext_block *block;
 	int count;
@@ -2985,7 +3165,7 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent)
 	if (all_blocks_are_justified_or_headers(ctx, *first_block))
 		return num_subtables;
 
-	gd.bounds = fz_infinite_rect;
+	gd.bounds = top_bounds;
 
 	/* First off, descend into any children to see if those look like tables. */
 	count = 0;
@@ -2995,7 +3175,10 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent)
 		{
 			if (block->u.s.down)
 			{
-				num_subtables += do_table_hunt(ctx, page, block->u.s.down);
+				int background = 0;
+				num_subtables += do_table_hunt(ctx, page, block->u.s.down, &background, top_bounds);
+				if (background)
+					*has_background = 1;
 				count++;
 			}
 		}
@@ -3009,6 +3192,11 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent)
 	if (count < 1)
 		return num_subtables;
 
+	/* If at least one of the subtables we found has a background on it, then
+	 * probably they all do, and we've probably got a good match. */
+	if (*has_background)
+		return num_subtables;
+
 	/* We only look for a table at this level if we either at the top
 	 * or on a div. This saves us looking for tables within an 'H'
 	 * for example. */
@@ -3026,6 +3214,8 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent)
 		if (find_table_within_bounds(ctx, &gd, *first_block, bounds))
 			break;
 
+		*has_background = gd.has_background;
+
 		if (num_subtables > 0)
 		{
 			/* We are risking throwing away a table we've already found for this
@@ -3096,12 +3286,31 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent)
 void
 fz_table_hunt(fz_context *ctx, fz_stext_page *page)
 {
+	int has_background = 0;
+
 	if (page == NULL)
 		return;
 
 	assert(verify_stext(ctx, page, NULL));
 
-	do_table_hunt(ctx, page, NULL);
+	do_table_hunt(ctx, page, NULL, &has_background, fz_infinite_rect);
+
+	assert(verify_stext(ctx, page, NULL));
+}
+
+void
+fz_table_hunt_within_bounds(fz_context *ctx, fz_stext_page *page, fz_rect rect)
+{
+	int has_background = 0;
+
+	if (page == NULL)
+		return;
+
+	assert(verify_stext(ctx, page, NULL));
+
+	fz_segment_stext_rect(ctx, page, rect);
+
+	do_table_hunt(ctx, page, NULL, &has_background, rect);
 
 	assert(verify_stext(ctx, page, NULL));
 }
