@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -34,19 +34,14 @@ typedef struct gstate
 	pdf_font_desc *font;
 } gstate;
 
-typedef struct resources_stack
-{
-	struct resources_stack *next;
-	pdf_obj *res;
-} resources_stack;
-
 typedef struct
 {
 	int num;
 	int gen;
-	int is_cff;
+	int is_ttf;
 	int is_cidfont;
 	pdf_obj *fontfile;
+	unsigned char digest[16];
 
 	fz_int_heap gids;
 	fz_int_heap cids;
@@ -67,7 +62,6 @@ typedef struct
 typedef struct
 {
 	pdf_processor super;
-	resources_stack *rstack;
 	fonts_usage_t *usage;
 	gstate *gs;
 } pdf_font_analysis_processor;
@@ -92,52 +86,10 @@ drop_processor(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_font_analysis_processor *p = (pdf_font_analysis_processor*)proc;
 
-	while (p->rstack)
-	{
-		resources_stack *stk = p->rstack;
-		p->rstack = stk->next;
-		pdf_drop_obj(ctx, stk->res);
-		fz_free(ctx, stk);
-	}
-
 	while (p->gs)
 		pop_gstate(ctx, p);
 }
 
-static void
-push_resources(fz_context *ctx, pdf_processor *proc, pdf_obj *res)
-{
-	pdf_font_analysis_processor *p = (pdf_font_analysis_processor *)proc;
-	resources_stack *stk = fz_malloc_struct(ctx, resources_stack);
-
-	stk->next = p->rstack;
-	p->rstack = stk;
-	fz_try(ctx)
-	{
-		stk->res = pdf_keep_obj(ctx, res);
-	}
-	fz_catch(ctx)
-	{
-		pdf_drop_obj(ctx, stk->res);
-		p->rstack = stk->next;
-		fz_free(ctx, stk);
-		fz_rethrow(ctx);
-	}
-}
-
-static pdf_obj *
-pop_resources(fz_context *ctx, pdf_processor *proc)
-{
-	pdf_font_analysis_processor *p = (pdf_font_analysis_processor *)proc;
-	resources_stack *stk = p->rstack;
-	pdf_obj *res = p->rstack->res;
-
-	p->rstack = stk->next;
-	fz_free(ctx, stk);
-
-	return res;
-}
-
 static void
 font_analysis_Q(fz_context *ctx, pdf_processor *proc)
 {
@@ -152,88 +104,102 @@ font_analysis_q(fz_context *ctx, pdf_processor *proc)
 	pdf_font_analysis_processor *p = (pdf_font_analysis_processor*)proc;
 	gstate *gs = p->gs;
 	gstate *new_gs = fz_malloc_struct(ctx, gstate);
-	*new_gs = *gs;
-	new_gs->next = gs;
 	p->gs = new_gs;
 
+	if (gs)
+	{
+		*new_gs = *gs;
+		new_gs->next = gs;
+	}
+
 	pdf_keep_font(ctx, new_gs->font);
+
 }
 
 static void
 font_analysis_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_font_desc *font, float size)
 {
 	pdf_font_analysis_processor *p = (pdf_font_analysis_processor*)proc;
-	pdf_obj *obj = pdf_dict_gets(ctx, pdf_dict_get(ctx, p->rstack->res, PDF_NAME(Font)), name);
+	pdf_obj *dict = pdf_lookup_resource(ctx, proc->rstack, PDF_NAME(Font), name);
 	pdf_obj *subtype, *fontdesc;
-	pdf_obj *key = NULL;
+	pdf_obj *fontfile = NULL;
+	pdf_obj *key;
 	int num, gen, i;
-	int is_cff = 0;
-	int cidfont = 0;
+	int is_cidfont = 0;
+	int is_ttf = 0;
+	unsigned char digest[16];
 
-	if (obj == NULL)
+	p->gs->current_font = -1; /* unknown font! */
+
+	if (dict == NULL)
 		return;
 
 	/* We can have multiple fonts that rely on the same underlying fontfile
 	 * object. Therefore, resolve down to that. */
-	subtype = pdf_dict_get(ctx, obj, PDF_NAME(Subtype));
+	subtype = pdf_dict_get(ctx, dict, PDF_NAME(Subtype));
 
-	if (pdf_name_eq(ctx, subtype, PDF_NAME(TrueType)))
+	if (subtype == PDF_NAME(Type1) || subtype == PDF_NAME(MMType1))
 	{
-		fontdesc = pdf_dict_get(ctx, obj, PDF_NAME(FontDescriptor));
-		key = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile2));
-		cidfont = 0;
+		// fontfile subtype should be Type1C for us to be able to subset it
+		key = PDF_NAME(FontFile);
+		fontdesc = pdf_dict_get(ctx, dict, PDF_NAME(FontDescriptor));
+		fontfile = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile));
+		is_cidfont = 0;
+		is_ttf = 0;
+	}
+	else if (subtype == PDF_NAME(TrueType))
+	{
+		key = PDF_NAME(FontFile2);
+		fontdesc = pdf_dict_get(ctx, dict, PDF_NAME(FontDescriptor));
+		fontfile = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile2));
+		is_cidfont = 0;
+		is_ttf = 1;
 	}
 	else if (pdf_name_eq(ctx, subtype, PDF_NAME(Type0)))
 	{
-		obj = pdf_array_get(ctx, pdf_dict_get(ctx, obj, PDF_NAME(DescendantFonts)), 0);
-		fontdesc = pdf_dict_get(ctx, obj, PDF_NAME(FontDescriptor));
-		key = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile2));
-		cidfont = 1; // fontsub7a
-		if (!key)
+		dict = pdf_array_get(ctx, pdf_dict_get(ctx, dict, PDF_NAME(DescendantFonts)), 0);
+		subtype = pdf_dict_get(ctx, dict, PDF_NAME(Subtype));
+		fontdesc = pdf_dict_get(ctx, dict, PDF_NAME(FontDescriptor));
+		if (subtype == PDF_NAME(CIDFontType0))
 		{
-			key = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile3));
-			subtype = pdf_dict_get(ctx, key, PDF_NAME(Subtype));
-			if (pdf_name_eq(ctx, subtype, PDF_NAME(OpenType)))
+			// fontfile subtype is either CIDFontType0C or OpenType
+			key = PDF_NAME(FontFile3);
+			fontfile = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile3));
+			subtype = pdf_dict_get(ctx, fontfile, PDF_NAME(Subtype));
+			if (subtype == PDF_NAME(CIDFontType0C))
 			{
-				cidfont = 0; // fontsub2a
+				is_cidfont = 1;
+				is_ttf = 0;
 			}
-			else if (pdf_name_eq(ctx, subtype, PDF_NAME(CIDFontType0C)))
+			else if (subtype == PDF_NAME(OpenType))
 			{
-				is_cff = 1;
+				is_cidfont = 1;
+				is_ttf = 1;
 			}
 			else
 			{
-				is_cff = 1;
-				cidfont = 0; // fontsub1a
+				fontfile = NULL;
 			}
 		}
-	}
-	else if (pdf_name_eq(ctx, subtype, PDF_NAME(Type1)))
-	{
-		fontdesc = pdf_dict_get(ctx, obj, PDF_NAME(FontDescriptor));
-		key = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile3));
-		is_cff = 1;
-		cidfont = 0;
-	}
-	else
-	{
-#ifdef DEBUG_SUBSETTING
-		fz_write_printf(ctx, fz_stddbg(ctx), "Unknown font of subtype ");
-		pdf_debug_obj(ctx, subtype);
-#endif
+		else if (subtype == PDF_NAME(CIDFontType2))
+		{
+			key = PDF_NAME(FontFile2);
+			fontfile = pdf_dict_get(ctx, fontdesc, PDF_NAME(FontFile2));
+			is_cidfont = 1;
+			is_ttf = 1;
+		}
 	}
 
-	if (!key)
+	if (!fontfile)
 	{
 #ifdef DEBUG_SUBSETTING
-		fz_write_printf(ctx, fz_stddbg(ctx), "No key found for font of subtype ");
-		pdf_debug_obj(ctx, subtype);
+		fz_write_printf(ctx, fz_stddbg(ctx), "No embedded file found for font of subtype %s\n", pdf_to_name(ctx, subtype));
 #endif
 		return;
 	}
 
-	num = pdf_to_num(ctx, key);
-	gen = pdf_to_gen(ctx, key);
+	num = pdf_to_num(ctx, fontfile);
+	gen = pdf_to_gen(ctx, fontfile);
 
 	for (i = 0; i < p->usage->len; i++)
 	{
@@ -242,6 +208,23 @@ font_analysis_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_fon
 			break;
 	}
 
+	fz_font_digest(ctx, font->font, digest);
+
+	/* Check for duplicate fonts. (Fonts in the document that have
+	 * the font stream included multiple times as different objects).
+	 * This can happen with naive insertion routines. */
+	if (i == p->usage->len)
+	{
+		for (i = 0; i < p->usage->len; i++)
+		{
+			if (memcmp(digest, p->usage->font[i].digest, 16) == 0)
+			{
+				pdf_dict_put(ctx, fontdesc, key, p->usage->font[i].fontfile);
+				break;
+			}
+		}
+	}
+
 	pdf_drop_font(ctx, p->gs->font);
 	p->gs->font = pdf_keep_font(ctx, font);
 	p->gs->current_font = i;
@@ -251,7 +234,7 @@ font_analysis_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_fon
 
 		for (j = 0; j < p->usage->font[i].len; j++)
 		{
-			if (pdf_objcmp(ctx, p->usage->font[i].font[j], obj) == 0)
+			if (pdf_objcmp(ctx, p->usage->font[i].font[j], dict) == 0)
 				return;
 		}
 
@@ -261,7 +244,7 @@ font_analysis_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_fon
 			p->usage->font[i].font = fz_realloc(ctx, p->usage->font[i].font, sizeof(*p->usage->font[i].font) * newmax);
 			p->usage->font[i].max = newmax;
 		}
-		p->usage->font[i].font[j] = pdf_keep_obj(ctx, obj);
+		p->usage->font[i].font[j] = pdf_keep_obj(ctx, dict);
 		p->usage->font[i].len++;
 
 		return;
@@ -277,9 +260,9 @@ font_analysis_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_fon
 		p->usage->max = n;
 	}
 
-	p->usage->font[i].is_cff = is_cff;
-	p->usage->font[i].is_cidfont = cidfont;
-	p->usage->font[i].fontfile = pdf_keep_obj(ctx, key);
+	p->usage->font[i].is_ttf = is_ttf;
+	p->usage->font[i].is_cidfont = is_cidfont;
+	p->usage->font[i].fontfile = pdf_keep_obj(ctx, fontfile);
 	p->usage->font[i].num = num;
 	p->usage->font[i].gen = gen;
 	p->usage->font[i].cids.len = 0;
@@ -291,12 +274,13 @@ font_analysis_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_fon
 	p->usage->font[i].len = 0;
 	p->usage->font[i].max = 0;
 	p->usage->font[i].font = NULL;
+	memcpy(p->usage->font[i].digest, digest, 16);
 	p->usage->len++;
 
 	p->usage->font[i].font = fz_malloc(ctx, sizeof(*p->usage->font[i].font) * 4);
 	p->usage->font[i].len = 1;
 	p->usage->font[i].max = 4;
-	p->usage->font[i].font[0] = pdf_keep_obj(ctx, obj);
+	p->usage->font[i].font[0] = pdf_keep_obj(ctx, dict);
 }
 
 static void
@@ -312,12 +296,14 @@ show_string(fz_context *ctx, pdf_font_analysis_processor *p, unsigned char *buf,
 	gstate *gs = p->gs;
 	pdf_font_desc *fontdesc = gs->font;
 	size_t pos = 0;
-	font_usage_t *font = &p->usage->font[gs->current_font];
+	font_usage_t *font;
 
-	/* e.g. for non-embedded base14 fonts. */
-	if (fontdesc == NULL)
+	// Not an embedded font!
+	if (gs->current_font < 0 || fontdesc == NULL)
 		return;
 
+	font = &p->usage->font[gs->current_font];
+
 	while (pos < len)
 	{
 		unsigned int cpt;
@@ -406,10 +392,6 @@ font_analysis_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pd
 	pdf_font_analysis_processor *pr = (pdf_font_analysis_processor *)proc;
 	pdf_document *doc = pdf_get_bound_document(ctx, xobj);
 	pdf_obj *resources = pdf_xobject_resources(ctx, xobj);
-
-	if (!resources)
-		resources = pr->rstack->res;
-
 	pdf_process_contents(ctx, (pdf_processor*)pr, doc, resources, xobj, NULL, NULL);
 }
 
@@ -419,8 +401,6 @@ pdf_new_font_analysis_processor(fz_context *ctx, fonts_usage_t *usage)
 	pdf_font_analysis_processor *proc = (pdf_font_analysis_processor *)pdf_new_processor(ctx, sizeof *proc);
 
 	proc->super.drop_processor = drop_processor;
-	proc->super.push_resources = push_resources;
-	proc->super.pop_resources = pop_resources;
 
 	proc->super.op_Do_form = font_analysis_Do_form;
 
@@ -441,6 +421,8 @@ pdf_new_font_analysis_processor(fz_context *ctx, fonts_usage_t *usage)
 		fz_rethrow(ctx);
 	}
 
+	proc->gs->current_font = -1; // no font set yet
+
 	proc->usage = usage;
 
 	return &proc->super;
@@ -452,14 +434,17 @@ examine_page(fz_context *ctx, pdf_document *doc, pdf_page *page, fonts_usage_t *
 	pdf_processor *proc = pdf_new_font_analysis_processor(ctx, usage);
 	pdf_obj *contents = pdf_page_contents(ctx, page);
 	pdf_obj *resources = pdf_page_resources(ctx, page);
-	pdf_annot *annot;
+	pdf_annot *annot, *widget;
 
 	fz_try(ctx)
 	{
 		pdf_process_contents(ctx, proc, doc, resources, contents, NULL, NULL);
 
+		pdf_processor_push_resources(ctx, proc, resources);
 		for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 			pdf_process_annot(ctx, proc, annot, NULL);
+		for (widget = pdf_first_widget(ctx, page); widget; widget = pdf_next_widget(ctx, widget))
+			pdf_process_annot(ctx, proc, widget, NULL);
 		pdf_close_processor(ctx, proc);
 	}
 	fz_always(ctx)
@@ -694,15 +679,32 @@ pdf_subset_fonts(fz_context *ctx, pdf_document *doc, int len, const int *pages)
 
 	fz_try(ctx)
 	{
-		/* Process every page. */
-		for (i = 0; i < len; i++)
+		if (len == 0)
 		{
-			page = pdf_load_page(ctx, doc, pages[i]);
+			/* Process every page. */
+			len = pdf_count_pages(ctx, doc);
+			for (i = 0; i < len; i++)
+			{
+				page = pdf_load_page(ctx, doc, i);
 
-			examine_page(ctx, doc, page, &usage);
+				examine_page(ctx, doc, page, &usage);
 
-			fz_drop_page(ctx, (fz_page *)page);
-			page = NULL;
+				fz_drop_page(ctx, (fz_page *)page);
+				page = NULL;
+			}
+		}
+		else
+		{
+			/* Process just the pages we are given. */
+			for (i = 0; i < len; i++)
+			{
+				page = pdf_load_page(ctx, doc, pages[i]);
+
+				examine_page(ctx, doc, page, &usage);
+
+				fz_drop_page(ctx, (fz_page *)page);
+				page = NULL;
+			}
 		}
 
 		/* All our font usage data is in heaps. Sort the heaps. */
@@ -741,10 +743,10 @@ pdf_subset_fonts(fz_context *ctx, pdf_document *doc, int len, const int *pages)
 			 * This will leave this font alone. */
 			fz_try(ctx)
 			{
-				if (font->is_cff)
-					subset_cff(ctx, doc, font, font->fontfile, symbolic, font->is_cidfont);
-				else
+				if (font->is_ttf)
 					subset_ttf(ctx, doc, font, font->fontfile, symbolic, font->is_cidfont);
+				else if (font->is_cidfont)
+					subset_cff(ctx, doc, font, font->fontfile, symbolic, font->is_cidfont);
 			}
 			fz_catch(ctx)
 			{
